"""Base jobs for DFTB+"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ase.calculators.dftb import Dftb

from quacc.runners.ase import Runner
from quacc.schemas.ase import Summarize
from quacc.utils.dicts import recursive_dict_merge

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms

    from quacc.types import Filenames, RunSchema, SourceDirectory

LOG_FILE = "dftb.out"
GEOM_FILE = "geo_end.gen"


def run_and_summarize(
    atoms: Atoms,
    calc_defaults: dict[str, Any] | None = None,
    calc_swaps: dict[str, Any] | None = None,
    additional_fields: dict[str, Any] | None = None,
    copy_files: SourceDirectory | dict[SourceDirectory, Filenames] | None = None,
) -> RunSchema:
    """
    Base job function for DFTB+ recipes.

    Parameters
    ----------
    atoms
        Atoms object
    calc_defaults
        The default calculator parameters to use.
    calc_swaps
        Dictionary of custom kwargs for the calculator that would override the
        calculator defaults. Set a value to `quacc.Remove` to remove a pre-existing key
        entirely. For a list of available keys, refer to the
        [ase.calculators.dftb.Dftb][] calculator.
    additional_fields
        Any additional fields to supply to the summarizer.
    copy_files
        Files to copy (and decompress) from source to the runtime directory.

    Returns
    -------
    RunSchema
        Dictionary of results, specified in [quacc.schemas.ase.Summarize.run][]
    """
    calc_flags = recursive_dict_merge(calc_defaults, calc_swaps)

    calc = Dftb(**calc_flags)
    final_atoms = Runner(atoms, calc, copy_files=copy_files).run_calc(
        geom_file=GEOM_FILE
    )

    return Summarize(additional_fields=additional_fields).run(final_atoms, atoms)



"""Core recipes for ORCA."""

from __future__ import annotations

from typing import TYPE_CHECKING

import psutil

from quacc import job
from quacc.atoms.core import perturb
from quacc.recipes.orca._base import run_and_summarize, run_and_summarize_opt

if TYPE_CHECKING:
    from typing import Any, Literal

    from ase.atoms import Atoms
    from numpy.typing import NDArray

    from quacc.types import Filenames, OptParams, OptSchema, RunSchema, SourceDirectory


@job
def static_job(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    xc: str = "wb97x-d3bj",
    basis: str = "def2-tzvp",
    orcasimpleinput: list[str] | None = None,
    orcablocks: list[str] | None = None,
    nprocs: int | Literal["max"] = "max",
    copy_files: SourceDirectory | dict[SourceDirectory, Filenames] | None = None,
    additional_fields: dict[str, Any] | None = None,
) -> RunSchema:
    """
    Carry out a single-point calculation.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    xc
        Exchange-correlation functional
    basis
        Basis set
    orcasimpleinput
        List of `orcasimpleinput` swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name. Refer to the
        [ase.calculators.orca.ORCA][] calculator for details on `orcasimpleinput`.
    orcablocks
        List of `orcablocks` swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name. Refer to the
        [ase.calculators.orca.ORCA][] calculator for details on `orcablocks`.
    nprocs
        Number of processors to use. Defaults to the number of physical cores.
    copy_files
        Files to copy (and decompress) from source to the runtime directory.
    additional_fields
        Additional fields to add to the results dictionary.

    Returns
    -------
    RunSchema
        Dictionary of results
    """
    additional_fields = {"name": "ORCA Static"} | (additional_fields or {})
    nprocs = psutil.cpu_count(logical=False) if nprocs == "max" else nprocs
    default_inputs = [xc, basis, "engrad", "normalprint"]
    default_blocks = [f"%pal nprocs {nprocs} end"]

    return run_and_summarize(
        atoms,
        charge,
        spin_multiplicity,
        default_inputs=default_inputs,
        default_blocks=default_blocks,
        input_swaps=orcasimpleinput,
        block_swaps=orcablocks,
        additional_fields=additional_fields,
        copy_files=copy_files,
    )


@job
def relax_job(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    xc: str = "wb97x-d3bj",
    basis: str = "def2-tzvp",
    run_freq: bool = False,
    orcasimpleinput: list[str] | None = None,
    orcablocks: list[str] | None = None,
    nprocs: int | Literal["max"] = "max",
    copy_files: SourceDirectory | dict[SourceDirectory, Filenames] | None = None,
    additional_fields: dict[str, Any] | None = None,
) -> RunSchema:
    """
    Carry out a geometry optimization.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    xc
        Exchange-correlation functional
    basis
        Basis set
    run_freq
        If a frequency calculation should be carried out.
    orcasimpleinput
        List of `orcasimpleinput` swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name. Refer to the
        [ase.calculators.orca.ORCA][] calculator for details on `orcasimpleinput`.
    orcablocks
        List of `orcablocks` swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name. Refer to the
        [ase.calculators.orca.ORCA][] calculator for details on `orcablocks`.
    nprocs
        Number of processors to use. Defaults to the number of physical cores.
    copy_files
        Files to copy (and decompress) from source to the runtime directory.
    additional_fields
        Additional fields to add to the results dictionary.

    Returns
    -------
    RunSchema
        Dictionary of results
    """
    additional_fields = {"name": "ORCA Relax"} | (additional_fields or {})
    nprocs = psutil.cpu_count(logical=False) if nprocs == "max" else nprocs

    default_inputs = [xc, basis, "normalprint", "opt"]
    if run_freq:
        default_inputs.append("freq")

    default_blocks = [f"%pal nprocs {nprocs} end"]

    return run_and_summarize(
        atoms,
        charge=charge,
        spin_multiplicity=spin_multiplicity,
        default_inputs=default_inputs,
        default_blocks=default_blocks,
        input_swaps=orcasimpleinput,
        block_swaps=orcablocks,
        additional_fields=additional_fields,
        copy_files=copy_files,
    )


@job
def freq_job(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    xc: str = "wb97x-d3bj",
    basis: str = "def2-tzvp",
    numerical: bool = False,
    orcasimpleinput: list[str] | None = None,
    orcablocks: list[str] | None = None,
    nprocs: int | Literal["max"] = "max",
    copy_files: SourceDirectory | dict[SourceDirectory, Filenames] | None = None,
    additional_fields: dict[str, Any] | None = None,
) -> RunSchema:
    """
    Carry out a vibrational frequency analysis calculation.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    xc
        Exchange-correlation functional
    basis
        Basis set
    numerical
        If True (default False), a numeric frequency calculation will be requested
    orcasimpleinput
        List of `orcasimpleinput` swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name. Refer to the
        [ase.calculators.orca.ORCA][] calculator for details on `orcasimpleinput`.
    orcablocks
        List of `orcablocks` swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name. Refer to the
        [ase.calculators.orca.ORCA][] calculator for details on `orcablocks`.
    nprocs
        Number of processors to use. Defaults to the number of physical cores.
    copy_files
        Files to copy (and decompress) from source to the runtime directory.
    additional_fields
        Additional fields to add to the results dictionary.

    Returns
    -------
    RunSchema
        Dictionary of results
    """
    nprocs = psutil.cpu_count(logical=False) if nprocs == "max" else nprocs
    default_inputs = [xc, basis, "normalprint", "numfreq" if numerical else "freq"]
    default_blocks = [f"%pal nprocs {nprocs} end"]

    return run_and_summarize(
        atoms,
        charge,
        spin_multiplicity,
        default_inputs=default_inputs,
        default_blocks=default_blocks,
        input_swaps=orcasimpleinput,
        block_swaps=orcablocks,
        additional_fields={"name": "ORCA Vibrational Frequency Analysis"}
        | (additional_fields or {}),
        copy_files=copy_files,
    )


@job
def ase_relax_job(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    xc: str = "wb97x-d3bj",
    basis: str = "def2-tzvp",
    orcasimpleinput: list[str] | None = None,
    orcablocks: list[str] | None = None,
    opt_params: OptParams | None = None,
    nprocs: int | Literal["max"] = "max",
    copy_files: SourceDirectory | dict[SourceDirectory, Filenames] | None = None,
    additional_fields: dict[str, Any] | None = None,
) -> OptSchema:
    """
    Carry out a geometry optimization.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    xc
        Exchange-correlation functional
    basis
        Basis set.
    orcasimpleinput
        List of `orcasimpleinput` swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name. Refer to the
        [ase.calculators.orca.ORCA][] calculator for details on `orcasimpleinput`.
    orcablocks
        List of `orcablocks` swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name. Refer to the
        [ase.calculators.orca.ORCA][] calculator for details on `orcablocks`.
    opt_params
        Dictionary of optimization parameters.
    nprocs
        Number of processors to use. Defaults to the number of physical cores.
    copy_files
        Files to copy (and decompress) from source to the runtime directory.
    additional_fields
        Additional fields to add to the results dictionary.

    Returns
    -------
    OptSchema
        Dictionary of results
    """
    nprocs = psutil.cpu_count(logical=False) if nprocs == "max" else nprocs
    default_inputs = [xc, basis, "engrad", "normalprint"]
    default_blocks = [f"%pal nprocs {nprocs} end"]

    return run_and_summarize_opt(
        atoms,
        charge=charge,
        spin_multiplicity=spin_multiplicity,
        default_inputs=default_inputs,
        default_blocks=default_blocks,
        input_swaps=orcasimpleinput,
        block_swaps=orcablocks,
        opt_params=opt_params,
        additional_fields={"name": "ORCA ASE Relax"} | (additional_fields or {}),
        copy_files=copy_files,
    )


@job
def ase_quasi_irc_job(
    atoms: Atoms,
    mode: list[list[float]] | NDArray,
    perturb_magnitude: float = 0.6,
    direction: Literal["forward", "reverse"] = "forward",
    charge: int = 0,
    spin_multiplicity: int = 1,
    xc: str = "wb97x-d3bj",
    basis: str = "def2-tzvp",
    orcasimpleinput: list[str] | None = None,
    orcablocks: list[str] | None = None,
    opt_params: OptParams | None = None,
    nprocs: int | Literal["max"] = "max",
    copy_files: SourceDirectory | dict[SourceDirectory, Filenames] | None = None,
    additional_fields: dict[str, Any] | None = None,
) -> OptSchema:
    """
    Quasi-IRC to optimize a reaction endpoint from a transition-state with known vibrational frequency modes.
    Perturbs the structure of `atoms` by a finite amount (0.6 * the normalized mode magnitude) along the specified
    vibrational frequency mode (assumed to be the transition mode), and then performs a `relax_job` on the perturbed
    structure.

    Parameters
    ----------
    atoms
        Atoms object
    mode
        Transition mode. This should be an Nx3 matrix, where N is the number of atoms in `atoms`.
    perturb_magnitude
        Factor to multiply the transition mode. Default is 0.6. In some cases, it may be advisable to increase this
        factor, perhaps to 1.0 or 1.1. Lowering it is not generally found to be helpful.
    direction
        Direction of the (Quasi)IRC. Should be "forward" or "reverse".
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    xc
        Exchange-correlation functional
    basis
        Basis set.
    orcasimpleinput
        List of `orcasimpleinput` swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name. Refer to the
        [ase.calculators.orca.ORCA][] calculator for details on `orcasimpleinput`.
    orcablocks
        List of `orcablocks` swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name. Refer to the
        [ase.calculators.orca.ORCA][] calculator for details on `orcablocks`.
    opt_params
        Dictionary of optimization parameters.
    nprocs
        Number of processors to use. Defaults to the number of physical cores.
    copy_files
        Files to copy (and decompress) from source to the runtime directory.
    additional_fields
        Additional fields to add to the results dictionary.

    Returns
    -------
    OptSchema
        Dictionary of results
    """
    nprocs = psutil.cpu_count(logical=False) if nprocs == "max" else nprocs
    default_inputs = [xc, basis, "engrad", "normalprint"]
    default_blocks = [f"%pal nprocs {nprocs} end"]

    scale = perturb_magnitude if direction == "forward" else perturb_magnitude * -1

    return run_and_summarize_opt(
        perturb(atoms, mode, scale),
        charge=charge,
        spin_multiplicity=spin_multiplicity,
        default_inputs=default_inputs,
        default_blocks=default_blocks,
        input_swaps=orcasimpleinput,
        block_swaps=orcablocks,
        opt_params=opt_params,
        additional_fields={"name": "ORCA ASE Quasi-IRC"} | (additional_fields or {}),
        copy_files=copy_files,
    )



"""Utility functions for running ASE calculators with ASE-based methods."""

from __future__ import annotations

from importlib.util import find_spec
from logging import getLogger
from shutil import copy, copytree
from typing import TYPE_CHECKING

import numpy as np
from ase.calculators import calculator
from ase.filters import FrechetCellFilter
from ase.io import Trajectory, read
from ase.md.md import MolecularDynamics
from ase.md.velocitydistribution import (
    MaxwellBoltzmannDistribution,
    Stationary,
    ZeroRotation,
)
from ase.optimize import BFGS
from ase.optimize.sciopt import SciPyOptimizer
from ase.vibrations import Vibrations
from monty.dev import requires
from monty.os.path import zpath

from quacc.atoms.core import copy_atoms
from quacc.runners._base import BaseRunner
from quacc.runners.prep import terminate
from quacc.utils.dicts import recursive_dict_merge

LOGGER = getLogger(__name__)

has_sella = bool(find_spec("sella"))


if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path
    from typing import Any

    from ase.atoms import Atoms
    from ase.calculators.calculator import Calculator
    from ase.optimize.optimize import Dynamics

    from quacc.types import (
        Filenames,
        MaxwellBoltzmanDistributionKwargs,
        SourceDirectory,
        VibKwargs,
    )


class Runner(BaseRunner):
    """
    Run various types of calculations in a scratch directory and copy the results back
    to the original directory. Note: This function does not modify the atoms object in-place.
    """

    def __init__(
        self,
        atoms: Atoms,
        calculator: Calculator,
        copy_files: SourceDirectory | dict[SourceDirectory, Filenames] | None = None,
    ) -> None:
        """
        Initialize the Runner object.

        Parameters
        ----------
        atoms
            The Atoms object to run calculations on.
        calculator
            The instantiated ASE calculator object to attach to the Atoms object.
        copy_files
            Files to copy (and decompress) from source to the runtime directory.

        Returns
        -------
        None
        """
        self.atoms = copy_atoms(atoms)
        self.atoms.calc = calculator
        self.copy_files = copy_files
        self.setup()

    def run_calc(
        self, geom_file: str | None = None, properties: list[str] | None = None
    ) -> Atoms:
        """
        This is a wrapper around `atoms.calc.calculate()`.

        Parameters
        ----------
        geom_file
            The filename of the log file that contains the output geometry, used to
            update the atoms object's positions and cell after a job. It is better
            to specify this rather than relying on ASE to update the positions, as the
            latter behavior varies between codes.
        properties
            List of properties to calculate. Defaults to ["energy"] if `None`.

        Returns
        -------
        Atoms
            The updated Atoms object.
        """
        if properties is None:
            properties = ["energy"]

        # Run calculation
        try:
            self.atoms.calc.calculate(self.atoms, properties, calculator.all_changes)
        except Exception as exception:
            terminate(self.tmpdir, exception)

        # Most ASE calculators do not update the atoms object in-place with a call
        # to .get_potential_energy(), which is important if an internal optimizer is
        # used. This section is done to ensure that the atoms object is updated to
        # the final geometry if `geom_file` is provided.
        # Note: We have to be careful to make sure we don't lose the calculator
        # object, as this contains important information such as the parameters
        # and output properties (e.g. final magnetic moments).
        if geom_file:
            atoms_new = read(zpath(str(self.tmpdir / geom_file)))

            # Make sure the atom indices didn't get updated somehow (sanity check).
            # If this happens, there is a serious problem.
            if (
                np.array_equal(
                    atoms_new.get_atomic_numbers(), self.atoms.get_atomic_numbers()
                )
                is False
            ):
                raise ValueError(
                    "Atomic numbers do not match between atoms and geom_file."
                )

            self.atoms.positions = atoms_new.positions
            self.atoms.cell = atoms_new.cell

        # Perform cleanup operations
        self.cleanup()

        return self.atoms

    def run_opt(
        self,
        relax_cell: bool = False,
        fmax: float | None = 0.01,
        max_steps: int = 1000,
        optimizer: Dynamics = BFGS,
        optimizer_kwargs: dict[str, Any] | None = None,
        store_intermediate_results: bool = False,
        fn_hook: Callable | None = None,
        run_kwargs: dict[str, Any] | None = None,
    ) -> Dynamics:
        """
        This is a wrapper around the optimizers in ASE.

        Parameters
        ----------
        relax_cell
            Whether to relax the unit cell shape and volume.
        fmax
            Tolerance for the force convergence (in eV/A).
        max_steps
            Maximum number of steps to take.
        optimizer
            Optimizer class to use.
        optimizer_kwargs
            Dictionary of kwargs for the optimizer. Takes all valid kwargs for ASE
            Optimizer classes. Refer to `_set_sella_kwargs` for Sella-related
            kwargs and how they are set.
        store_intermediate_results
            Whether to store the files generated at each intermediate step in the
            optimization. If enabled, they will be stored in a directory named
            `stepN` where `N` is the step number, starting at 0.
        fn_hook
            A custom function to call after each step of the optimization.
            The function must take the instantiated dynamics class as
            its only argument.
        run_kwargs
            Dictionary of kwargs for the `run()` method of the optimizer.

        Returns
        -------
        Dynamics
            The ASE Dynamics object following an optimization.
        """
        # Set defaults
        merged_optimizer_kwargs = recursive_dict_merge(
            {"logfile": self.tmpdir / "opt.log", "restart": self.tmpdir / "opt.json"},
            optimizer_kwargs,
        )
        run_kwargs = run_kwargs or {}
        traj_filename = "opt.traj"

        # Check if trajectory kwarg is specified
        if "trajectory" in merged_optimizer_kwargs:
            msg = "Quacc does not support setting the `trajectory` kwarg."
            raise ValueError(msg)

        # Handle optimizer kwargs
        if (
            issubclass(optimizer, SciPyOptimizer | MolecularDynamics)
            or optimizer.__name__ == "IRC"
        ):
            # https://gitlab.com/ase/ase/-/issues/1476
            # https://gitlab.com/ase/ase/-/merge_requests/3310
            merged_optimizer_kwargs.pop("restart", None)
        if optimizer.__name__ == "Sella":
            self._set_sella_kwargs(merged_optimizer_kwargs)

        # Define the Trajectory object
        traj_file = self.tmpdir / traj_filename
        traj = Trajectory(traj_file, "w", atoms=self.atoms)
        merged_optimizer_kwargs["trajectory"] = traj

        # Set volume relaxation constraints, if relevant
        if relax_cell and self.atoms.pbc.any():
            self.atoms = FrechetCellFilter(self.atoms)

        # Run optimization
        full_run_kwargs = {"fmax": fmax, "steps": max_steps, **run_kwargs}
        if issubclass(optimizer, MolecularDynamics):
            full_run_kwargs.pop("fmax")
        try:
            with traj, optimizer(self.atoms, **merged_optimizer_kwargs) as dyn:
                if issubclass(optimizer, SciPyOptimizer | MolecularDynamics):
                    # https://gitlab.coms/ase/ase/-/issues/1475
                    # https://gitlab.com/ase/ase/-/issues/1497
                    dyn.run(**full_run_kwargs)
                else:
                    for i, _ in enumerate(dyn.irun(**full_run_kwargs)):
                        if store_intermediate_results:
                            self._copy_intermediate_files(
                                i,
                                files_to_ignore=[
                                    traj_file,
                                    merged_optimizer_kwargs.get("restart"),
                                    merged_optimizer_kwargs.get("logfile"),
                                ],
                            )
                        if fn_hook:
                            fn_hook(dyn)
        except Exception as exception:
            terminate(self.tmpdir, exception)

        # Perform cleanup operations
        self.cleanup()
        traj.filename = zpath(str(self.job_results_dir / traj_filename))
        dyn.trajectory = traj

        return dyn

    def run_vib(self, vib_kwargs: VibKwargs | None = None) -> Vibrations:
        """
        Run an ASE-based vibration analysis in a scratch directory and copy the results back
        to the original directory. This can be useful if file I/O is slow in the working
        directory, so long as file transfer speeds are reasonable.

        This is a wrapper around the vibrations module in ASE.

        Parameters
        ----------
        vib_kwargs
            Dictionary of kwargs for the [ase.vibrations.Vibrations][] class.

        Returns
        -------
        Vibrations
            The updated Vibrations module
        """
        # Set defaults
        vib_kwargs = vib_kwargs or {}

        # Run calculation
        vib = Vibrations(self.atoms, name=str(self.tmpdir / "vib"), **vib_kwargs)
        try:
            vib.run()
        except Exception as exception:
            terminate(self.tmpdir, exception)

        # Summarize run
        vib.summary(log=str(self.tmpdir / "vib_summary.log"))

        # Perform cleanup operations
        self.cleanup()

        return vib

    def run_md(
        self,
        dynamics: MolecularDynamics,
        dynamics_kwargs: dict[str, Any] | None = None,
        steps: int = 1000,
        maxwell_boltzmann_kwargs: MaxwellBoltzmanDistributionKwargs | None = None,
        set_com_stationary: bool = False,
        set_zero_rotation: bool = False,
    ) -> MolecularDynamics:
        """
        Run an ASE-based MD in a scratch directory and copy the results back to
        the original directory.

        Parameters
        ----------
        dynamics
            MolecularDynamics class to use, from `ase.md.md.MolecularDynamics`.
        dynamics_kwargs
            Dictionary of kwargs for the dynamics. Takes all valid kwargs for ASE
            MolecularDynamics classes.
        steps
            Maximum number of steps to run
        maxwell_boltzmann_kwargs
            If specified, a `MaxwellBoltzmannDistribution` will be applied to the atoms
            based on `ase.md.velocitydistribution.MaxwellBoltzmannDistribution` with the
            specified keyword arguments.
        set_com_stationary
            Whether to set the center-of-mass momentum to zero. This would be applied after
            any `MaxwellBoltzmannDistribution` is set.
        set_zero_rotation
            Whether to set the total angular momentum to zero. This would be applied after
            any `MaxwellBoltzmannDistribution` is set.

        Returns
        -------
        MolecularDymamics
            The ASE MolecularDynamics object.
        """

        # Set defaults
        dynamics_kwargs = dynamics_kwargs or {}
        maxwell_boltzmann_kwargs = maxwell_boltzmann_kwargs or {}
        dynamics_kwargs["logfile"] = self.tmpdir / "md.log"

        if maxwell_boltzmann_kwargs:
            MaxwellBoltzmannDistribution(self.atoms, **maxwell_boltzmann_kwargs)
        if set_com_stationary:
            Stationary(self.atoms)
        if set_zero_rotation:
            ZeroRotation(self.atoms)

        return self.run_opt(
            fmax=None,
            max_steps=steps,
            optimizer=dynamics,
            optimizer_kwargs=dynamics_kwargs,
        )

    def _copy_intermediate_files(
        self, step_number: int, files_to_ignore: list[Path] | None = None
    ) -> None:
        """
        Copy all files in the working directory to a subdirectory named `stepN` where `N`
        is the step number. This is useful for storing intermediate files generated during
        an ASE relaaxation.

        Parameters
        ----------
        step_number
            The step number.
        files_to_ignore
            A list of files to ignore when copying files to the subdirectory.

        Returns
        -------
        None
        """
        files_to_ignore = files_to_ignore or []
        store_path = self.tmpdir / f"step{step_number}"
        store_path.mkdir()
        for item in self.tmpdir.iterdir():
            if not item.name.startswith("step") and item not in files_to_ignore:
                if item.is_file():
                    copy(item, store_path)
                elif item.is_dir():
                    copytree(item, store_path / item.name)

    @requires(has_sella, "Sella must be installed. Refer to the quacc documentation.")
    def _set_sella_kwargs(self, optimizer_kwargs: dict[str, Any]) -> None:
        """
        Modifies the `optimizer_kwargs` in-place to address various Sella-related
        parameters. This function does the following for the specified key/value pairs in
        `optimizer_kwargs`:

        1. Sets `order = 0` if not specified (i.e. minimization rather than TS
        by default).

        2. If `internal` is not defined and not `atoms.pbc.any()`, set it to `True`.

        Parameters
        ----------
        optimizer_kwargs
            The kwargs for the Sella optimizer.

        Returns
        -------
        None
        """
        if "order" not in optimizer_kwargs:
            optimizer_kwargs["order"] = 0

        if not self.atoms.pbc.any() and "internal" not in optimizer_kwargs:
            optimizer_kwargs["internal"] = True



"""Core recipes for Psi4."""

from __future__ import annotations

from importlib.util import find_spec
from typing import TYPE_CHECKING

from monty.dev import requires

from quacc import job
from quacc.recipes.psi4._base import run_and_summarize

has_psi4 = bool(find_spec("psi4"))

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms

    from quacc.types import Filenames, RunSchema, SourceDirectory


@job
@requires(has_psi4, "Psi4 not installed. Try conda install -c psi4 psi4")
def static_job(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    method: str = "wb97x-v",
    basis: str = "def2-tzvp",
    copy_files: SourceDirectory | dict[SourceDirectory, Filenames] | None = None,
    additional_fields: dict[str, Any] | None = None,
    **calc_kwargs,
) -> RunSchema:
    """
    Function to carry out a single-point calculation.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    method
        The level of theory to use.
    basis
        Basis set
    copy_files
        Files to copy (and decompress) from source to the runtime directory.
    additional_fields
        Additional fields to add to the results dictionary.
    **calc_kwargs
        Custom kwargs for the Psi4 calculator. Set a value to
        `quacc.Remove` to remove a pre-existing key entirely. For a list of available
        keys, refer to the [ase.calculators.psi4.Psi4][] calculator.

    Returns
    -------
    RunSchema
        Dictionary of results from [quacc.schemas.ase.Summarize.run][].
        See the type-hint for the data structure.
    """
    calc_defaults = {
        "mem": "16GB",
        "num_threads": "max",
        "method": method,
        "basis": basis,
        "charge": charge,
        "multiplicity": spin_multiplicity,
        "reference": "uks" if spin_multiplicity > 1 else "rks",
    }
    return run_and_summarize(
        atoms,
        charge=charge,
        spin_multiplicity=spin_multiplicity,
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        additional_fields={"name": "Psi4 Static"} | (additional_fields or {}),
        copy_files=copy_files,
    )



"""
Custom types used throughout quacc.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic_settings import BaseSettings


class DefaultSetting(BaseSettings):
    """
    Type hint for when a default setting will be applied
    """


if TYPE_CHECKING:
    from collections.abc import Callable
    from datetime import datetime
    from pathlib import Path
    from typing import Any, Literal

    from ase.atoms import Atoms
    from ase.md.md import MolecularDynamics
    from ase.optimize.optimize import Dynamics
    from emmet.core.math import ListMatrix3D, Matrix3D, Vector3D
    from emmet.core.symmetry import CrystalSystem
    from emmet.core.vasp.calc_types import CalcType
    from emmet.core.vasp.calc_types.enums import RunType, TaskType
    from emmet.core.vasp.calculation import VaspObject
    from emmet.core.vasp.task_valid import TaskState
    from numpy.random import Generator
    from numpy.typing import ArrayLike, NDArray
    from pymatgen.core.composition import Composition
    from pymatgen.core.lattice import Lattice
    from pymatgen.core.periodic_table import Element
    from pymatgen.core.structure import Molecule, Structure
    from pymatgen.entries.computed_entries import ComputedEntry
    from pymatgen.io.vasp.inputs import Incar, Kpoints, Poscar, Potcar
    from typing_extensions import NotRequired, TypedDict

    # ----------- File handling -----------

    Filenames = str | Path | list[str | Path]
    SourceDirectory = str | Path

    # ----------- k-point handling -----------

    class PmgKpts(TypedDict, total=False):
        """
        Type hint for `pmg_kpts` in [quacc.utils.kpts.convert_pmg_kpts][].
        """

        line_density: float
        kppvol: float
        kppa: float
        length_densities: tuple[float, float, float]

    # ----------- Runner parameter type hints -----------

    class OptParams(TypedDict, total=False):
        """
        Type hint for `opt_params` used throughout quacc.
        """

        relax_cell: bool
        fmax: float | None
        max_steps: int
        optimizer: Dynamics
        optimizer_kwargs: dict[str, Any] | None
        store_intermediate_results: bool
        fn_hook: Callable | None
        run_kwargs: dict[str, Any] | None

    class MDParams(TypedDict, total=False):
        """
        Type hint for `md_params` used throughout quacc.
        """

        dynamics: MolecularDynamics
        dynamics_kwargs: dict[str, Any] | None
        steps: int
        maxwell_boltzmann_kwargs: MaxwellBoltzmanDistributionKwargs | None
        set_com_stationary: bool
        set_zero_rotation: bool

    class VibKwargs(TypedDict, total=False):
        """
        Type hint for `vib_kwargs` in [quacc.runners.ase.Runner.run_vib][].
        """

        indices: list[int] | None
        delta: float
        nfree: int

    class MaxwellBoltzmanDistributionKwargs(TypedDict, total=False):
        """
        Type hint for `maxwell_boltzmann_kwargs` in [quacc.runners.ase.Runner.run_md][].
        """

        temperature_K: float
        force_temp: bool
        rng: Generator | None

    # ----------- Atoms handling type hints -----------

    class AdsSiteFinderKwargs(TypedDict, total=False):
        """
        Type hint for `ads_site_finder_kwargs` in [quacc.atoms.slabs.make_adsorbate_structures][].
        """

        selective_dynamics: bool  # default = False
        height: float  # default = 0.9
        mi_vec: ArrayLike | None  # default = None

    class FindAdsSitesKwargs(TypedDict, total=False):
        """
        Type hint for `find_ads_sites_kwargs` in [quacc.atoms.slabs.make_adsorbate_structures][].
        """

        distance: float  # default = 2.0
        put_inside: bool  # default = True
        symm_reduce: float  # default = 1e-2
        near_reduce: float  # default = 1e-2
        positions: list[
            Literal["ontop", "bridge", "hollow", "subsurface"]
        ]  # default: ["ontop", "bridge", "hollow"]
        no_obtuse_hollow: bool  # default = True

    # ----------- Custom calculator type hints -----------

    class QchemResults(TypedDict, total=False):
        """
        Type hint for the `results` attribute in [quacc.calculators.qchem.qchem.QChem][].
        """

        energy: float  # electronic energy in eV
        taskdoc: dict[str, Any]  # Output from `emmet.core.qc_tasks.TaskDoc`
        hessian: NotRequired[NDArray]  # Hessian in eV/A^2/amu
        forces: NotRequired[NDArray]  # forces in eV/A

    class VaspJobKwargs(TypedDict, total=False):
        """
        Type hint for `vasp_job_kwargs` in in [quacc.calculators.vasp.vasp_custodian.run_custodian][].
        """

        output_file: str  # default = "vasp.out"
        stderr_file: str  # default = "std_err.txt"
        suffix: str  # default = ""
        final: bool  # default = True
        backup: bool  # default = True
        auto_npar: bool  # default = False
        auto_gamma: bool  # default = True
        settings_override: dict | None  # default = None
        copy_magmom: bool  # default = False
        auto_continue: bool  # default = False

    class VaspCustodianKwargs(TypedDict, total=False):
        """
        Type hint for `custodian_kwargs` in [quacc.calculators.vasp.vasp_custodian.run_custodian][].
        """

        max_errors_per_job: int | None  # default = None
        polling_time_step: int  # default = 10
        monitor_freq: int  # default = 10
        skip_over_errors: bool  # default = False
        gzipped_output: bool  # default = False
        checkpoint: bool  # default = False
        terminate_func: Callable | None  # default = None
        terminate_on_nonzero_returncode: bool  # default = False

    # ----------- ASE calculator type hints -----------

    class Results(TypedDict):
        """Dictionary of results from atoms.calc.results"""

    class Parameters(TypedDict):
        """Dictionary of parameters from atoms.calc.parameters"""

    # ----------- ASE dynamics type hints -----------

    class ParametersDyn(TypedDict):
        """Dictionary of parameters from Dynamics.todict()"""

    class TrajectoryLog(TypedDict):
        """Dictionary of parameters related to the MD trajectory"""

        # ASE units
        kinetic_energy: float
        temperature: float
        time: float

    # ----------- Emmet type hints -----------

    class SymmetryData(TypedDict):
        """Type hint associated with [emmet.core.symmetry.SymmetryData][]"""

        crystal_system: CrystalSystem
        symbol: str
        number: int
        point_group: str
        symprec: float
        version: str

    class PointGroupData(TypedDict):
        """Type hint associated with [emmet.core.symmetry.PointGroupData][]"""

        point_group: str
        rotation_number: float
        linear: bool
        tolerance: float
        eigen_tolerance: float
        matrix_tolerance: float

    class EmmetBase(TypedDict):
        """Type hint associated with `emmet.core.base.EmmetBaseModel`."""

        emmet_version: str
        pymatgen_version: str
        pull_request: int | None
        database_version: str | None
        build_date: datetime
        license: Literal["BY-C", "BY-NC"]

    class StructureMetadata(EmmetBase):
        """Type hint associated with [emmet.core.structure.StructureMetadata][]"""

        nsites: int
        elements: list[Element]
        nelements: int
        composition: Composition
        formula_pretty: str
        formula_anonymous: str
        chemsys: str
        volume: float
        density: float
        density_atomic: float
        symmetry: SymmetryData

    class MoleculeMetadata(EmmetBase):
        """Type hint associated with [emmet.core.structure.MoleculeMetadata][]"""

        charge: int
        spin_multiplicity: int
        natoms: int
        elements: list[Element]
        nelements: int
        nelectrons: int
        composition: Composition
        composition_reduced: Composition
        formula_alphabetical: str
        formula_pretty: str
        formula_anonymous: str
        chemsys: str
        symmetry: PointGroupData

    class PotcarSpec(TypedDict):
        """Type hint associated with emmet.core.vasp.calculation.PotcarSpec."""

        titel: str
        hash: str

    class CalculationInput(TypedDict):
        """Type hint associated with emmet.core.vasp.calculation.CalculationInput."""

        incar: dict[str, Any]
        kpoints: dict[str, Any]
        nkpoints: int
        potcar: list[str]
        potcar_spec: list[PotcarSpec]
        potcar_type: list[str]
        parameters: dict[str, Any]
        lattice_rec: Lattice
        structure: Structure
        is_hubbard: bool
        hubbards: dict[str, float]

    class ElectronicStep(TypedDict):
        """Type hint associated with emmet.core.vasp.calculation.ElectronicStep."""

        alphaZ: float
        ewald: float
        hartreedc: float
        XCdc: float
        pawpsdc: float
        pawaedc: float
        eentropy: float
        bandstr: float
        atom: float
        e_fr_energy: float
        e_wo_entrp: float
        e_0_energy: float

    class IonicStep(TypedDict):
        """Type hint associated with emmet.core.vasp.calculation.IonicStep."""

        e_fr_energy: float
        e_wo_entrp: float
        e_0_energy: float
        forces: list[Vector3D]
        stress: Matrix3D
        electronic_steps: list[ElectronicStep]
        structure: Structure

    class FrequencyDependentDielectric(TypedDict):
        """Type hint associated with
        emmet.core.vasp.calculation.FrequencyDependentDielectric.
        """

        real: list[list[float]]
        imaginary: list[list[float]]
        energy: list[float]

    class ElectronPhononDisplacedStructures(TypedDict):
        """Type hint associated with
        emmet.core.vasp.calculation.ElectronPhononDisplacedStructures.
        """

        temperatures: list[float]
        structures: list[Structure]

    class RunStatistics(TypedDict):
        """Type hint associated with emmet.core.vasp.calculation.RunStatistics."""

        average_memory: float
        max_memory: float
        elapsed_time: float
        system_time: float
        user_time: float
        total_time: float
        cores: int

    class CalculationOutput(TypedDict, total=False):
        """Type hint associated with emmet.core.vasp.calculation.CalculationOutput."""

        energy: float
        energy_per_atom: float
        structure: Structure
        efermi: float
        is_metal: bool
        bandgap: float
        cbm: float
        vbm: float
        is_gap_direct: bool
        direct_gap: float
        transition: str
        mag_density: float
        epsilon_static: ListMatrix3D
        epsilon_static_wolfe: ListMatrix3D
        epsilon_ionic: ListMatrix3D
        frequency_dependent_dielectric: FrequencyDependentDielectric
        ionic_steps: list[IonicStep]
        locpot: dict[int, list[float]]
        outcar: dict[str, Any]
        force_constants: list[list[Matrix3D]]
        normalmode_frequencies: list[float]
        normalmode_eigenvals: list[float]
        normalmode_eigenvecs: list[list[Vector3D]]
        elph_displaced_structures: ElectronPhononDisplacedStructures
        dos_properties: dict[str, dict[str, dict[str, float]]]
        run_stats: RunStatistics

    class Calculation(TypedDict):
        """Type hint associated with emmet.core.vasp.calculation.Calculation."""

        dir_name: str
        vasp_version: str
        has_vasp_completed: TaskState | bool
        input: CalculationInput
        output: CalculationOutput
        completed_at: str
        task_name: str
        output_file_paths: dict[str, str]
        bader: dict
        run_type: RunType
        task_type: TaskType
        calc_type: CalcType

    class OrigInputs(TypedDict):
        """Type hint associated with emmet.core.tasks.OrigInputs."""

        incar: Incar
        poscar: Poscar
        kpoints: Kpoints
        potcar: Potcar

    class OutputDoc(TypedDict):
        """Type hint associated with emmet.core.tasks.OutputDoc."""

        structure: Structure
        density: float
        energy: float
        forces: list[list[float]]
        stress: list[list[float]]
        energy_per_atom: float
        bandgap: float

    class CustodianDoc(TypedDict):
        """Type hint associated with emmet.core.tasks.CustodianDoc."""

        corrections: list[Any]
        job: dict

    class AnalysisDoc(TypedDict):
        """Type hint associated with emmet.core.tasks.AnalysisDoc."""

        delta_volume: float
        delta_volume_percent: float
        max_force: float
        warnings: list[str]
        errors: list[str]

    class TaskDoc(StructureMetadata):
        """Type hint associated with emmet.core.tasks.TaskDoc."""

        tags: list[str] | None
        dir_name: str
        state: TaskState
        calcs_reversed: list[Calculation]
        structure: Structure
        task_type: CalcType | TaskType
        task_id: str
        orig_inputs: OrigInputs
        output: OutputDoc
        included_objects: VaspObject
        vasp_objects: dict[VaspObject, Any]
        entry: ComputedEntry
        task_label: str
        author: str
        transformations: dict[str, Any]
        additional_json: dict[str, Any]
        custodian: list[CustodianDoc]
        analysis: AnalysisDoc
        last_updated: datetime

    # ----------- Schema (Atoms) type hints -----------

    class AtomsSchema(StructureMetadata, MoleculeMetadata):
        """Type hint associated with [quacc.schemas.atoms.atoms_to_metadata][]"""

        atoms: Atoms
        structure: Structure  # if atoms.pbc.any()
        molecule: Molecule  # if not atoms.pbc.any()

    # ----------- Schema (ASE) type hints -----------

    class RunSchema(AtomsSchema):
        """Schema for [quacc.schemas.ase.Summarize.run][]"""

        input_atoms: AtomsSchema | None
        nid: str
        dir_name: str
        parameters: Parameters
        results: Results
        quacc_version: str

    class OptSchema(RunSchema):
        """Schema for [quacc.schemas.ase.Summarize.opt][]"""

        parameters_opt: ParametersDyn
        converged: bool
        trajectory: list[Atoms]
        trajectory_results: list[Results]

    class DynSchema(RunSchema):
        """Schema for [quacc.schemas.ase.Summarize.md][]"""

        parameters_md: ParametersDyn
        trajectory: list[Atoms]
        trajectory_log: TrajectoryLog
        trajectory_results: list[Results]

    class ParametersVib(TypedDict):
        delta: float
        direction: str
        method: str
        ndof: int
        nfree: int

    class VibResults(TypedDict):
        imag_vib_freqs: int
        n_imag: int
        vib_energies: list[float]
        vib_freqs: list[float]
        vib_energies_raw: list[float]
        vib_freqs_raw: list[float]

    class VibSchema(AtomsSchema):
        parameters: Parameters | None
        parameters_vib: ParametersVib | None
        results: VibResults

    class ParametersThermo(TypedDict):
        # ASE units
        temperature: float
        pressure: float
        sigma: int
        spin_multiplicity: int
        vib_freqs: list[float]
        vib_energies: list[float]
        n_imag: int

    class ThermoResults(TypedDict):
        # ASE units
        energy: float
        enthalpy: float
        entropy: float
        gibbs_energy: float
        zpe: float

    class ThermoSchema(AtomsSchema):
        parameters_thermo: ParametersThermo
        results: ThermoResults

    class VibThermoSchema(VibSchema, ThermoSchema):
        """Combined Vibrations and Thermo schema"""

    # ----------- Schema (phonons) type hints -----------

    class ThermalProperties(TypedDict):
        """Type hint associated with PhononSchema."""

        temperatures: NDArray
        free_energy: NDArray
        entropy: NDArray
        heat_capacity: NDArray

    class MeshProperties(TypedDict):
        """Type hint associated with PhononSchema."""

        qpoints: NDArray
        weights: NDArray
        frequencies: NDArray
        eigenvectors: NDArray
        group_velocities: NDArray

    class DosProperties(TypedDict):
        """Type hint associated with PhononSchema."""

        frequency_points: NDArray
        total_dos: NDArray

    class PhononResults(TypedDict):
        thermal_properties: ThermalProperties
        mesh_properties: MeshProperties
        total_dos: DosProperties
        force_constants: NDArray

    class PhonopyMetadata(TypedDict):
        """Type hint associated with PhononSchema."""

        version: str

    class PhononSchema(AtomsSchema):
        """Type hint associated with [quacc.schemas.phonons.summarize_phonopy][]"""

        parameters: dict[str, Any] | None
        nid: str
        dir_name: str
        phonopy_metadata: PhonopyMetadata
        results: PhononResults
        quacc_version: str

    # ----------- Schema (VASP) type hints -----------

    class BaderSchema(TypedDict, total=False):
        """Type hint associated with quacc.schemas.vasp._bader_runner."""

        atomic_volume: float
        partial_charges: list[float]
        spin_moments: list[float]
        bader_version: float
        min_dist: list[float]

    class DDECSchema(TypedDict, total=False):
        """Type hint associated with quacc.schemas.vasp._ddec_runner."""

        partial_charges: list[float]
        spin_moments: list[float]
        dipoles: list[float]
        bond_order_sums: list[float]
        bond_order_dict: dict
        rsquared_moments: list[float]
        rcubed_moments: list[float]
        rfourth_moments: list[float]

    class CM5Schema(TypedDict):
        """Type hint used in DDECSchema"""

        partial_charges: list[float]

    class ChargemolSchema(TypedDict, total=False):
        """Type hint associated with quacc.schemas.vasp._chargemol_runner`"""

        ddec: DDECSchema
        cm5: CM5Schema

    class VaspSchema(RunSchema, TaskDoc):
        """Type hint associated with [quacc.schemas.vasp.VaspSummarize.run][]"""

        bader: BaderSchema
        chargemol: ChargemolSchema
        steps: dict[int, TaskDoc]  # when store_intermediate_results=True

    # ----------- Recipe (VASP) type hints -----------

    class DoubleRelaxSchema(TypedDict):
        """Type hint associated with the double relaxation jobs."""

        relax1: VaspSchema
        relax2: VaspSchema

    class MPGGARelaxFlowSchema(VaspSchema):
        """Type hint associated with the MP GGA relaxation flows."""

        relax1: VaspSchema
        relax2: VaspSchema
        static: VaspSchema

    class MPMetaGGARelaxFlowSchema(MPGGARelaxFlowSchema):
        """Type hint associated with the MP meta-GGA relaxation flows."""

        prerelax: VaspSchema

    class QMOFRelaxSchema(VaspSchema):
        """Type hint associated with the QMOF relaxation jobs."""

        prerelax_lowacc: VaspSchema | None
        position_relax_lowacc: VaspSchema
        volume_relax_lowacc: VaspSchema | None
        double_relax: VaspSchema

    class VaspASEOptSchema(VaspSchema, OptSchema):
        """Type hint associated with VASP relaxations run via ASE"""

    # ----------- Recipe (Espresso) type hints -----------
    class SystemData(TypedDict):
        occupations: str
        smearing: str
        degauss: float

    class ElectronsData(TypedDict):
        conv_thr: float
        mixing_mode: str
        mixing_beta: float

    class InputData(TypedDict):
        system: SystemData
        electrons: ElectronsData
        control: NotRequired[dict[str, Any]]

    class EspressoBaseSet(TypedDict):
        input_data: InputData
        kspacing: float

    class EspressoBandsSchema(TypedDict, total=False):
        bands_pw: RunSchema
        bands_pp: RunSchema
        fermi_surface: RunSchema

    class EspressoDosSchema(TypedDict):
        static_job: RunSchema
        non_scf_job: RunSchema
        dos_job: RunSchema

    class EspressoProjwfcSchema(TypedDict):
        static_job: RunSchema
        non_scf_job: RunSchema

    class EspressoPhononDosSchema(TypedDict):
        phonon_job: RunSchema
        q2r_job: RunSchema
        matdyn_job: RunSchema

    # ----------- Recipe (NewtonNet) type hints -----------

    class NewtonNetTSSchema(OptSchema):
        freq_job: VibThermoSchema | None

    class NewtonNetIRCSchema(OptSchema):
        freq_job: VibThermoSchema | None

    class NewtonNetQuasiIRCSchema(OptSchema):
        irc_job: NewtonNetIRCSchema
        freq_job: VibThermoSchema | None

    # ----------- Recipe (Q-Chem) type hints -----------

    class QchemQuasiIRCSchema(OptSchema):
        initial_irc: OptSchema



"""Schemas for VASP."""

from __future__ import annotations

import os
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING

from ase.io import read
from emmet.core.tasks import TaskDoc
from monty.os.path import zpath
from pymatgen.command_line.bader_caller import bader_analysis_from_path
from pymatgen.command_line.chargemol_caller import ChargemolAnalysis
from pymatgen.entries.compatibility import (
    CompatibilityError,
    MaterialsProject2020Compatibility,
)

from quacc import QuaccDefault, get_settings
from quacc.atoms.core import get_final_atoms_from_dynamics
from quacc.schemas.ase import Summarize
from quacc.utils.dicts import finalize_dict, recursive_dict_merge

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms
    from ase.optimize.optimize import Optimizer
    from maggma.core import Store

    from quacc.types import (
        BaderSchema,
        ChargemolSchema,
        DefaultSetting,
        VaspASEOptSchema,
        VaspSchema,
    )


LOGGER = getLogger(__name__)


class VaspSummarize:
    """
    Get tabulated results from a VASP run and store them in a database-friendly format.
    """

    def __init__(
        self,
        directory: str | Path | None = None,
        move_magmoms: bool = True,
        run_bader: bool | DefaultSetting = QuaccDefault,
        run_chargemol: bool | DefaultSetting = QuaccDefault,
        check_convergence: bool | DefaultSetting = QuaccDefault,
        report_mp_corrections: bool = False,
        additional_fields: dict[str, Any] | None = None,
    ) -> None:
        """
        Initialize the VASP summarizer.

        Parameters
        ----------
        directory
            Path to VASP outputs. A value of None specifies the calculator directory.
        move_magmoms
            Whether to move the final magmoms of the original Atoms object to the
            initial magmoms of the returned Atoms object.
        run_bader
            Whether a Bader analysis should be performed. Will not run if bader
            executable is not in PATH even if bader is set to True. Defaults to
            VASP_BADER in settings.
        run_chargemol
            Whether a Chargemol analysis should be performed. Will not run if chargemol
            executable is not in PATH even if chargmeol is set to True. Defaults to
            VASP_CHARGEMOL in settings.
        check_convergence
            Whether to throw an error if convergence is not reached. Defaults to True in
            settings.
        report_mp_corrections
            Whether to apply the MP corrections to the task document. Defaults to False.
        additional_fields
            Additional fields to add to the task document.

        Returns
        -------
        None
        """
        self.directory = directory
        self.move_magmoms = move_magmoms
        self.run_bader = run_bader
        self.run_chargemol = run_chargemol
        self.check_convergence = check_convergence
        self.report_mp_corrections = report_mp_corrections
        self.additional_fields = additional_fields or {}
        self._settings = get_settings()

    def run(
        self, final_atoms: Atoms, store: Store | None | DefaultSetting = QuaccDefault
    ) -> VaspSchema:
        """
        Get tabulated results from a VASP run and store them in a database-friendly format.

        Parameters
        ----------
        final_atoms
            ASE Atoms object following a calculation.
        store
            Maggma Store object to store the results in. Defaults to `QuaccSettings.STORE`,

        Returns
        -------
        VaspSchema
            Dictionary representation of the task document
        """
        run_bader = (
            self._settings.VASP_BADER
            if self.run_bader == QuaccDefault
            else self.run_bader
        )
        run_chargemol = (
            self._settings.VASP_CHARGEMOL
            if self.run_chargemol == QuaccDefault
            else self.run_chargemol
        )
        check_convergence = (
            self._settings.CHECK_CONVERGENCE
            if self.check_convergence == QuaccDefault
            else self.check_convergence
        )
        directory = Path(self.directory or final_atoms.calc.directory)
        store = self._settings.STORE if store == QuaccDefault else store
        additional_fields = self.additional_fields or {}

        # Fetch all tabulated results from VASP outputs files. Fortunately, emmet
        # already has a handy function for this
        vasp_task_model = TaskDoc.from_directory(directory)

        # Get MP corrections
        if self.report_mp_corrections:
            mp_compat = MaterialsProject2020Compatibility()
            try:
                corrected_entry = mp_compat.process_entry(
                    vasp_task_model.structure_entry, on_error="raise"
                )
                vasp_task_model.entry = corrected_entry
            except CompatibilityError as err:
                LOGGER.warning(err)

        # Convert the VASP task model to a dictionary
        vasp_task_doc = vasp_task_model.model_dump()

        # Check for calculation convergence
        if check_convergence and vasp_task_doc["state"] != "successful":
            raise RuntimeError(
                f"VASP calculation did not converge. Will not store task data. Refer to {directory}"
            )
        poscar_path = directory / "POSCAR"
        initial_atoms = read(zpath(str(poscar_path)))
        base_task_doc = Summarize(
            directory=directory,
            move_magmoms=self.move_magmoms,
            additional_fields=self.additional_fields,
        ).run(final_atoms, initial_atoms, store=None)

        if nsteps := len([f for f in os.listdir(directory) if f.startswith("step")]):
            intermediate_vasp_task_docs = {
                "steps": {
                    n: TaskDoc.from_directory(directory / f"step{n}").model_dump()
                    for n in range(nsteps)
                    if (directory / f"step{n}").is_dir()
                }
            }
        else:
            intermediate_vasp_task_docs = {}

        # Get Bader analysis
        if run_bader:
            try:
                bader_results = bader_runner(directory)
            except Exception:
                bader_results = None
                LOGGER.warning("Bader analysis could not be performed.", exc_info=True)

            if bader_results:
                vasp_task_doc["bader"] = bader_results

        # Get the Chargemol analysis
        if run_chargemol:
            try:
                chargemol_results = chargemol_runner(directory)
            except Exception:
                chargemol_results = None
                LOGGER.warning(
                    "Chargemol analysis could not be performed.", exc_info=True
                )

            if chargemol_results:
                vasp_task_doc["chargemol"] = chargemol_results

        # Make task document
        unsorted_task_doc = (
            intermediate_vasp_task_docs
            | vasp_task_doc
            | base_task_doc
            | additional_fields
        )
        return finalize_dict(
            unsorted_task_doc,
            directory=directory,
            gzip_file=self._settings.GZIP_FILES,
            store=store,
        )

    def ase_opt(
        self,
        optimizer: Optimizer,
        trajectory: list[Atoms] | None = None,
        store: Store | None | DefaultSetting = QuaccDefault,
    ) -> VaspASEOptSchema:
        """
        Summarize an ASE-based VASP optimization.

        Parameters
        ----------
        optimizer
            The ASE optimizer object
        trajectory
            ASE Trajectory object or list[Atoms] from reading a trajectory file. If
            None, the trajectory must be found in dyn.trajectory.filename.
        store
            Maggma Store object to store the results in. Defaults to `QuaccSettings.STORE`,
        """
        store = self._settings.STORE if store == QuaccDefault else store

        final_atoms = get_final_atoms_from_dynamics(optimizer)
        directory = Path(self.directory or final_atoms.calc.directory)

        opt_run_summary = Summarize(
            directory=directory,
            move_magmoms=self.move_magmoms,
            additional_fields=self.additional_fields,
        ).opt(
            optimizer,
            trajectory=trajectory,
            check_convergence=self.check_convergence,
            store=None,
        )

        vasp_summary = self.run(final_atoms, store=None)
        unsorted_task_doc = recursive_dict_merge(vasp_summary, opt_run_summary)
        return finalize_dict(
            unsorted_task_doc,
            directory=directory,
            gzip_file=self._settings.GZIP_FILES,
            store=store,
        )


def bader_runner(path: Path | str) -> BaderSchema:
    """
    Runs a Bader partial charge and spin moment analysis using the VASP output
    files in the given path. This function requires that `bader` is located in
    your PATH environment variable. See
    http://theory.cm.utexas.edu/henkelman/code/bader for the bader code.

    Parameters
    ----------
    path
        The path where the VASP output files are located. Must include CHGCAR,
        AECCAR0, AECCAR2, and POTCAR files. These files can be gzip'd or not --
        it doesn't matter. If None, the current working directory is used.

    Returns
    -------
    BaderSchema
        Dictionary containing the Bader analysis summary
    """
    # Make sure files are present
    relevant_files = ["AECCAR0", "AECCAR2", "CHGCAR", "POTCAR"]
    for f in relevant_files:
        if not Path(path, f).exists() and not Path(path, f"{f}.gz").exists():
            msg = f"Could not find {f} in {path}."
            raise FileNotFoundError(msg)

    bader_stats = bader_analysis_from_path(path)

    # Store the partial charge, which is much more useful than the raw charge
    # and is more intuitive than the charge transferred. An atom with a positive
    # partial charge is cationic, whereas an atom with a negative partial charge
    # is anionic.
    bader_stats["partial_charges"] = [-c for c in bader_stats["charge_transfer"]]

    # Some cleanup of the returned dictionary
    if "magmom" in bader_stats:
        bader_stats["spin_moments"] = bader_stats["magmom"]
    for k in ["charge", "charge_transfer", "reference_used", "magmom"]:
        bader_stats.pop(k, None)

    return bader_stats


def chargemol_runner(
    path: Path | str, atomic_densities_path: str | None = None
) -> ChargemolSchema:
    """
    Runs a Chargemol (i.e. DDEC6 + CM5) analysis using the VASP output files in
    the given path. This function requires that the chargemol executable, given
    by the name `Chargemol_09_26_2017_linux_parallel`,
    `Chargemol_09_26_2017_linux_serial`, or `chargemol` is in the system PATH
    environment variable. See https://sourceforge.net/projects/ddec/files for
    the Chargemol code.

    Parameters
    ----------
    path
        The path where the VASP output files are located. Must include CHGCAR,
        AECCAR0, AECCAR2, and POTCAR files. These files can be gzip'd or not --
        it doesn't matter. If None, the current working directory is used.
    atomic_densities_path
        The path where the reference atomic densities are located for Chargemol.
        If None, we assume that this directory is defined in an environment
        variable named DDEC6_ATOMIC_DENSITIES_DIR. See the Chargemol
        documentation for more information.

    Returns
    -------
    ChargemolSchema
        Dictionary containing the Chargemol analysis summary
    """
    # Make sure files are present
    relevant_files = ["AECCAR0", "AECCAR2", "CHGCAR", "POTCAR"]
    for f in relevant_files:
        if not Path(path, f).exists() and not Path(path, f"{f}.gz").exists():
            msg = f"Could not find {f} in {path}."
            raise FileNotFoundError(msg)

    # Check environment variable
    if atomic_densities_path is None and "DDEC6_ATOMIC_DENSITIES_DIR" not in os.environ:
        msg = "DDEC6_ATOMIC_DENSITIES_DIR environment variable not defined."
        raise OSError(msg)

    # Run Chargemol analysis
    return ChargemolAnalysis(path=path, atomic_densities_path=atomic_densities_path)



"""Utility functions for interfacing with databases."""

from __future__ import annotations

import uuid
from typing import TYPE_CHECKING

from monty.json import jsanitize

if TYPE_CHECKING:
    from typing import Any

    from maggma.core import Store


def results_to_db(store: Store, results: dict[str, Any] | list[dict]) -> None:
    """
    Store the results of a quacc recipe in a user-specified Maggma Store. A UUID will be
    generated for each entry.

    Parameters
    ----------
    store
        The Maggma Store object to store the results in
    results
        The output summary dictionary or list of dictionaries from a quacc
        recipe

    Returns
    -------
    None
    """
    if not isinstance(results, list):
        results = [results]

    sanitized_results = [
        jsanitize(result, enum_values=True, recursive_msonable=True)
        for result in results
    ]

    for result in sanitized_results:
        result["uuid"] = str(uuid.uuid4())

    with store:
        store.update(sanitized_results, key="uuid")



"""Common slab workflows."""

from __future__ import annotations

from typing import TYPE_CHECKING

from quacc import subflow
from quacc.atoms.slabs import make_adsorbate_structures, make_slabs_from_bulk

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms

    from quacc import Job


@subflow
def bulk_to_slabs_subflow(
    atoms: Atoms,
    relax_job: Job,
    static_job: Job | None = None,
    make_slabs_kwargs: dict[str, Any] | None = None,
) -> list[dict]:
    """
    Workflow consisting of:

    1. Slab generation

    2. Slab relaxations

    3. Slab statics (optional)

    Parameters
    ----------
    atoms
        Atoms object
    relax_job
        The relaxation function.
    static_job
        The static function.
    make_slabs_kwargs
        Additional keyword arguments to pass to
        [quacc.atoms.slabs.make_slabs_from_bulk][]

    Returns
    -------
    list[dict]
        List of schemas.
    """
    make_slabs_kwargs = make_slabs_kwargs or {}

    slabs = make_slabs_from_bulk(atoms, **make_slabs_kwargs)

    results = []
    for slab in slabs:
        result = relax_job(slab)

        if static_job is not None:
            result = static_job(result["atoms"])

        results.append(result)

    return results


@subflow
def slab_to_ads_subflow(
    atoms: Atoms,
    adsorbate: Atoms,
    relax_job: Job,
    static_job: Job | None,
    make_ads_kwargs: dict[str, Any] | None = None,
) -> list[dict]:
    """
    Workflow consisting of:

    1. Slab-adsorbate generation

    2. Slab-adsorbate relaxations

    3. Slab-adsorbate statics (optional)

    Parameters
    ----------
    atoms
        Atoms object for the slab structure.
    adsorbate
        Atoms object for the adsorbate.
    relax_job
        The slab releaxation job.
    static_job
        The slab static job.
    make_ads_kwargs
        Additional keyword arguments to pass to
        [quacc.atoms.slabs.make_adsorbate_structures][]

    Returns
    -------
    list[dict]
        List of schemas.
    """
    make_ads_kwargs = make_ads_kwargs or {}

    slabs = make_adsorbate_structures(atoms, adsorbate, **make_ads_kwargs)

    results = []
    for slab in slabs:
        result = relax_job(slab)

        if static_job is not None:
            result = static_job(result["atoms"])

        results.append(result)

    return results
